@@ -1,17 +1,18 @@
 /*
+ *  lft_ifname.c
+ *  Layer Four Traceroute
+ *
+ *  This file is part of LFT.
+ *
+ *  The LFT software provided in this Distribution is
+ *  Copyright 2007 VOSTROM Holdings, Inc.
+ *
+ *  The full text of our legal notices is contained in the file called
+ *  COPYING, included with this Distribution.
+ *
+ */
 
-  This file is part of LFT.
-
-  The LFT software provided in this Distribution is
-  Copyright 2004 VOSTROM Holdings, Inc.
- 
-  The full text of our legal notices is contained in the file called
-  COPYING, included with this Distribution.
-
-  ifname.c : ljd
-  $Id: lft_ifname.c,v 1.7 2004/11/26 21:20:10 vic Exp $
-*/
-
+#if !defined(WIN32) && !defined(_WIN32)
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
@@ -26,14 +27,19 @@
 #include <unistd.h>
 
 #include "lft_ifname.h"
-#include "acconfig.h"
+#if defined( __CYGWIN__ ) || defined( WIN32 ) || defined(_WIN32)
+#include "config/acconfig.win.h"
+#else
+#include "config/acconfig.h"
+#endif
 
-static int		sock = -1;
+static int sock = -1;
 
 u_long
 lft_getifaddr (const char *ifname)
 {
-	struct ifreq		ifr;
+	struct ifreq ifr;
+	struct sockaddr_in addr;
 
 	/* Only do this once of course */
 	if (sock < 0) {
@@ -43,14 +49,22 @@ lft_getifaddr (const char *ifname)
 		}
 	}
 
-	STRNCPY(ifr.ifr_name, (char*)ifname, IFNAMSIZ);
+	STRNCPY(ifr.ifr_name, ifname, IFNAMSIZ);
 
 	if (ioctl(sock, SIOCGIFADDR, &ifr) < 0) {
 		perror("ioctl");
 		return -1;
 	}
 
-	return ((struct sockaddr_in *)(&(ifr.ifr_addr)))->sin_addr.s_addr;
+	if (ifr.ifr_addr.sa_family != AF_INET) {
+		fprintf (stderr, "%s: Interface not configured with IPv4 address.\n", ifname);
+		fflush (stderr);
+		return -1;
+	}
+
+	memcpy(&addr, &ifr.ifr_addr, sizeof addr);
+
+	return (addr.sin_addr.s_addr);
 }
 
 char *
@@ -91,7 +105,7 @@ lft_getifname (struct in_addr addr)
 
 		if (ifr.ifr_addr.sa_family != AF_INET) continue;
 
-		thisaddr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
+		thisaddr = ((const struct sockaddr_in *)(const void *)(&(ifr.ifr_addr)))->sin_addr;
 
 		if (thisaddr.s_addr == addr.s_addr)
 		  return strdup(ifr.ifr_name);
@@ -131,3 +145,4 @@ main (int argc, char *argv[])
 }
 
 #endif /*LFT_IFNAME_TESTING*/
+#endif
