@@ -20,8 +20,8 @@
 *       cc -o whob whois.c -DSTANDALONE
 *
 *   
+*   Portions (c) Victor Oppleman (lft@oppleman.com)
 *   Portions (c) 2011 Markus Gothe <nietzsche@lysator.liu.se>
-*   Portions (c) 2007 Victor Oppleman (lft@oppleman.com)
 *   Portions (c) 2002 Ugen Antsilevitch (ugen@xonix.com)
 *
 */
@@ -48,6 +48,7 @@
 /* OPTIONS and variable initialization */
 
 static char pwhois_server[] = "whois.pwhois.org";
+static char myaddress_server[] = "myaddress.today";
 static char radb_server[] = "whois.ra.net";
 static char cymru_server[] = "whois.cymru.com";
 static char arin_server[] = "whois.arin.net";
@@ -56,8 +57,9 @@ static char ripe_server[] = "whois.ripe.net";
 static char ripe_ris_server[] = "riswhois.ripe.net";
 
 #ifdef STANDALONE
-const char *version = "3.7";                    /* set version string for library and client */
-const char *appname = "WhoB";            	/* set application name */
+const char *version = "3.8";                   /* set version string for library and client */
+const char *appname = "WhoB";            	    /* set application name */
+static int go_interactive = 0;                  /* We'll wait on STDIN unless args suggest otherwise */
 static int use_cymru = 0;                       /* Don't use Cymru by default */
 static int use_riswhois = 0;                    /* Don't use RIPE NCC RIS by default */
 static int display_orgname = 1;                 /* Display orgname by default */
@@ -65,6 +67,7 @@ static int display_aspath = 0;                  /* Don't display AS-PATH by defa
 static int display_netname = 0;                 /* Don't display netname by default */
 static int display_radb_as = 0;                 /* Don't display RADB Origin-AS by default */
 static int show_routes_byasn = 0;               /* Don't show all advertisements by default */
+static int show_routes_bytransitasn = 0;        /* Don't show all routes that transit ASN by default */
 static int show_networks_byasn = 0;             /* Don't show all networks by default */
 static int show_contacts_byasn = 0;             /* Don't show all contact info by default */
 static int show_routes_byprefix = 0;            /* Don't show all routes by prefix by default */
@@ -74,22 +77,13 @@ static int read_from_file = 0;                  /* Don't read input from file by
 static int riswhoisfromfile = 0;                /* Don't use riswhois by default for readfromfile */
 static int cymrufromfile = 0;                   /* Don't use Cymru by default for readfromfile */
 static int use_gigo = 1;                        /* Use GIGO feature by default */
+static int use_stdin = 0;                       /* Don't use STDIN for bulk file input by default */
 static const unsigned int max_hostname_input = 200;      /* Maximum length of input from user */
-static const int max_lines = 1000;              /* Maximum lines to read from bulk file per query */
+static const int max_lines = 2500;              /* Maximum lines to read from bulk file per query */
 static const int line_size = 256;               /* Maximum line length */
 static char hostname[256];
 #endif
 
-/* Optimize DNS resolution */
-/*
-    #include <resolv.h>
-    extern struct __res_state _res;
-
-    res_init();
-    _res.retrans = 1;
-    _res.retry = 1;
-*/
-
 /* END of OPTIONS and variable initialization */
 #if defined(WIN32) || defined(_WIN32)
 char *
@@ -339,15 +333,47 @@ static char *match_inparens(char *target)
 
 static char *match_afterparens(char *target)
 {
-    /* matches something like "   (HELLO) xxx" and returns a pointer to "xxx" */
-    target = strchr(target, '(');
-    if (!target) return NULL;
-    target = strchr(target, ')');
-    if (!target) return NULL;
-    target++;
-    while(*target && isspace(*target)) target++;
-    if (*target) return strdup(target);
-    else return NULL;
+	/* matches something like "   (HELLO) xxx" and returns a pointer to "xxx" */
+	target = strchr(target, '(');
+	if (!target) return NULL;
+	target = strchr(target, ')');
+	if (!target) return NULL;
+	target++;
+	while(*target && isspace(*target)) target++;
+	if (*target) return strdup(target);
+	else return NULL;
+}
+
+static int stricontains(const char *haystack, const char *needle) 
+{
+	/* Search for a substring without a string library */
+	int i, j, match;
+	i = 0, j = 0;
+	
+	while (haystack[i] != '\0') {
+	
+	  while (tolower(haystack[i]) != tolower(needle[0]) && haystack[i] != '\0')
+		 i++;
+	
+	  if (haystack[i] == '\0')
+		 return (-1);
+	
+	  match = i;
+	
+	  while (tolower(haystack[i]) == tolower(needle[j]) && haystack[i] != '\0' && needle[j] != '\0') {
+		 i++;
+		 j++;
+	  }
+	
+	  if (needle[j] == '\0')
+		 return (match);
+	  if (haystack[i] == '\0')
+		 return (-1);
+	
+	  i = match + 1;
+	  j = 0;
+	}
+	return 0;
 }
 
 whois_session_params *w_init(void)
@@ -765,7 +791,7 @@ w_lookup_all_riswhois_bulk(whois_session_params * wsess, struct ip_list_array *i
                 if(wsess->logprintfCookie)
                     lft_printf(wsess->logprintfCookie,"Skipping additional object for same query.\n");
                 else
-                    printf("Skipping additional object for same query.\n");
+                    fprintf(stderr,"Skipping additional object for same query.\n");
             }
         } else
         if ((value = match_prefix("descr:", responses[i].ptr))) {
@@ -978,7 +1004,7 @@ w_lookup_as_cymru_bulk(whois_session_params * wsess, struct ip_list_array *iplis
                     if(wsess->logprintfCookie)
                         lft_printf(wsess->logprintfCookie,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                     else
-                        printf("Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
+                        fprintf(stderr,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                 }
                 break;
             }
@@ -1040,6 +1066,48 @@ static int w_display_rvbyasn_pwhois(whois_session_params * wsess, char *asn)
     return 0;
 }
 
+static int w_display_rvbytransitasn_pwhois(whois_session_params * wsess, char *asn)
+{
+    char *reply;
+    const char *query_begin = "routeview transit-as=";
+    char *whob_query = NULL;
+    char *serv;
+    
+    if (!asn) 
+        return -1;
+    
+    if (strlen(wsess->pw_serv) > 0)
+        serv = wsess->pw_serv;
+    else
+        serv = pwhois_server;
+    
+    whob_query = malloc( (strlen(appname) + strlen(version) + strlen(query_begin) + strlen(asn)) * sizeof(char) + 10);
+    whob_query[0]=0;
+    
+    /* prepare the text-string-based query */
+    strcat(whob_query,"app=\"");
+    strcat(whob_query,appname);
+    strcat(whob_query," ");
+    strcat(whob_query,version);
+    strcat(whob_query,"\" ");
+    strcat(whob_query,query_begin);
+    strcat(whob_query,asn);
+    
+    strcat(whob_query,"\n");
+    
+    reply = w_ask(serv, whob_query, NULL);
+    if (!reply) {
+        if (wsess->w_noisy) fprintf(stderr,"No reply from %s.\n",serv);
+        return -1;
+    }
+    
+    printf("%s",reply);
+    
+    free(reply); free(whob_query);
+    
+    return 0;
+}
+
 static int w_display_contactsbyasn_pwhois(whois_session_params * wsess, char *asn)
 {
     char *reply;
@@ -1192,7 +1260,7 @@ static int w_display_bulkfromfile_pwhois(whois_session_params * wsess, char *fil
         exit(EXIT_FAILURE);
     }  
     
-    if (!strncmp(filespec,"-",1)) {
+    if (!strncmp(filespec,"-",1) || use_stdin == 1) {
         bulkFile = stdin;
     } else 
         bulkFile = fopen(filespec, "r");
@@ -1300,7 +1368,7 @@ static int w_display_bulkfromfile_riswhois(whois_session_params * wsess, char *f
         exit(EXIT_FAILURE);
     }
 
-    if (!strncmp(filespec,"-",1)) {
+    if (!strncmp(filespec,"-",1) || use_stdin == 1) {
         bulkFile = stdin;
     } else
         bulkFile = fopen(filespec, "r");
@@ -1399,7 +1467,7 @@ static int w_display_bulkfromfile_cymru(whois_session_params * wsess, char *file
         exit(EXIT_FAILURE);
     }
 
-    if (!strncmp(filespec,"-",1)) {
+    if (!strncmp(filespec,"-",1) || use_stdin == 1) {
         bulkFile = stdin;
     } else
         bulkFile = fopen(filespec, "r");
@@ -1485,7 +1553,7 @@ static int w_display_pwhois_version(whois_session_params * wsess)
         serv = pwhois_server;
     
     if (wsess->w_noisy)
-        printf("Querying '%s' for version/status.\n",serv);
+        fprintf(stderr,"Querying '%s' for version/status.\n",serv);
     
     reply = w_ask(serv, "version", NULL);
     if (!reply) {
@@ -1500,6 +1568,58 @@ static int w_display_pwhois_version(whois_session_params * wsess)
     return 0;
 }
 
+static int w_display_myaddress_http(whois_session_params * wsess)
+{
+    char *reply;
+    char *serv;
+
+    if (strlen(wsess->pw_serv) > 0)
+        serv = wsess->pw_serv;
+    else
+        serv = myaddress_server;
+
+    if (wsess->w_noisy)
+        fprintf(stderr,"Querying '%s' for my public address.\n",serv);
+
+    reply = w_ask(serv, "Accept:*/*\nHost: myaddress.today\nUser-Agent: WhoB\nGET /\n\n", "80");
+    if (!reply) {
+        if(wsess->w_noisy)
+                fprintf(stderr,"No reply from %s.\n",serv);
+        return -1;
+    }
+  
+    printf("%s",reply);
+
+    free(reply);
+    return 0;
+}
+
+static int w_display_myaddress(whois_session_params * wsess)
+{
+    char *reply;
+    char *serv;
+
+    if (strlen(wsess->pw_serv) > 0)
+        serv = wsess->pw_serv;
+    else
+        serv = pwhois_server;
+
+    if (wsess->w_noisy)
+        fprintf(stderr,"Querying '%s' for my public address.\n",serv);
+
+    reply = w_ask(serv, "whoami", NULL);
+    if (!reply) {
+        if(wsess->w_noisy)
+                fprintf(stderr,"No reply from %s.\n",serv);
+        return -1;
+    }
+   
+    printf("%s",reply);
+
+    free(reply);
+    return 0;
+}
+
 static int w_display_pwhois_gigo(whois_session_params * wsess, char *user_query)
 {
     char *reply;
@@ -1511,7 +1631,7 @@ static int w_display_pwhois_gigo(whois_session_params * wsess, char *user_query)
         serv = pwhois_server;
     
     if (wsess->w_noisy)
-        printf("Querying '%s' for: '%s'\n",serv,user_query);
+        fprintf(stderr,"Querying '%s' for: '%s'\n",serv,user_query);
     
     reply = w_ask(serv, user_query, NULL);
     if (!reply) {
@@ -1543,7 +1663,7 @@ w_lookup_all_pwhois_bulk(whois_session_params * wsess, struct ip_list_array *ipl
     if (!iplist)
         return -1;
     
-    if ((*iplist).application) {
+    if ((uintptr_t)(*iplist).application) {
         bulk_ip_query = (char *)malloc(((strlen((*iplist).application) * sizeof(char)) +10) + ((strlen(bulk_begin) + strlen(bulk_end) + 1) * sizeof(char)) + (16 * (*iplist).numItems));
     } else 
 	bulk_ip_query = (char *)malloc(((strlen(appname) * sizeof(char)) +10) + ((strlen(version) + strlen(bulk_begin) + strlen(bulk_end) + 1) * sizeof(char)) + (16 * (*iplist).numItems));
@@ -1558,7 +1678,7 @@ w_lookup_all_pwhois_bulk(whois_session_params * wsess, struct ip_list_array *ipl
     
     /* prepare the text-string-based query */
     strcat(bulk_ip_query,bulk_begin);
-    if ((*iplist).application) {
+    if ((uintptr_t)(*iplist).application) {
         strcat(bulk_ip_query, "app=\"");
         strcat(bulk_ip_query, (*iplist).application);
         strcat(bulk_ip_query, "\"\n");
@@ -1611,7 +1731,7 @@ w_lookup_all_pwhois_bulk(whois_session_params * wsess, struct ip_list_array *ipl
                         if(wsess->logprintfCookie)
                             lft_printf(wsess->logprintfCookie,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                         else
-                            printf("Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
+                            fprintf(stderr,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                     }
                     break;
                 }
@@ -1663,7 +1783,7 @@ w_lookup_all_pwhois_bulk_ext(whois_session_params * wsess, struct ext_ip_list_ar
         return -1;
     iplist->geoavailable=0;
 
-    if ((*iplist).application) {
+    if ((uintptr_t)(*iplist).application) {
         bulk_ip_query = malloc(((strlen((*iplist).application) * sizeof(char)) +10) + ((strlen(bulk_begin) + strlen(bulk_end) + 1) * sizeof(char)) + (16 * (*iplist).numItems));
     } else bulk_ip_query = malloc(((strlen(appname) * sizeof(char)) +10) + ((strlen(version) + strlen(bulk_begin) + strlen(bulk_end) + 1) * sizeof(char)) + (16 * (*iplist).numItems));
     bulk_ip_query[0]=0;
@@ -1677,7 +1797,7 @@ w_lookup_all_pwhois_bulk_ext(whois_session_params * wsess, struct ext_ip_list_ar
     
     /* prepare the text-string-based query */
     strcat(bulk_ip_query,bulk_begin);
-    if ((*iplist).application) {
+    if ((uintptr_t)(*iplist).application) {
         strcat(bulk_ip_query,"app=\"");
         strcat(bulk_ip_query,(*iplist).application);
         strcat(bulk_ip_query,"\"\n");
@@ -1727,7 +1847,7 @@ w_lookup_all_pwhois_bulk_ext(whois_session_params * wsess, struct ext_ip_list_ar
                         if(wsess->logprintfCookie)
                             lft_printf(wsess->logprintfCookie,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                         else
-                            printf("Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
+                            fprintf(stderr,"Parse error at \'%c\': non-numeric value at position %d of %s).\n",value[i],i,value);
                     }
                     break;
                 }
@@ -1768,7 +1888,7 @@ w_lookup_all_pwhois_bulk_ext(whois_session_params * wsess, struct ext_ip_list_ar
                         if(wsess->logprintfCookie)
                             lft_printf(wsess->logprintfCookie,"Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
                         else
-                            printf("Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
+                            fprintf(stderr,"Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
                     }
                     break;
                 }
@@ -1787,7 +1907,7 @@ w_lookup_all_pwhois_bulk_ext(whois_session_params * wsess, struct ext_ip_list_ar
                         if(wsess->logprintfCookie)
                             lft_printf(wsess->logprintfCookie,"Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
                         else
-                            printf("Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
+                           fprintf(stderr,"Parse error at \'%c\': can't parse value at position %d of %s).\n",value[i],i,value);
                     }
                     break;
                 }
@@ -2268,6 +2388,7 @@ usage (char *prog)
              "  -c         Use Cymru\'s whois server instead of pWhoIs\n"
              "  -r         Use RIPE NCC\'s RIS whois server instead of pWhoIs\n"
              "\nPrefix WhoIs Advanced Options:\n"
+             "  -A         Display all routes transiting the target ASN (pWhoIs-only)\n"
              "  -a         Display all routes announced by the target ASN (pWhoIs-only)\n"
              "  -P         Display all routes respective to a target prefix (pWhoIs-only)\n"
              "  -N         Display all networks registered to the target ASN (pWhoIs-only)\n"
@@ -2276,6 +2397,7 @@ usage (char *prog)
              "  -s         Display the version and status of the pWhoIs server (pWhoIs-only)\n"
              "  -V         Display verbose/debug output.  Use more \'V\'s for additional detail\n"
              "  -v         Display WhoB\'s version information and exit\n"  
+             "  me|whoami  Display the public IP address of this host\n"
              "\n"
              ,version,prog);
     fprintf(stderr,"Example:     %s -gnp 1.2.3.4\n",prog);
@@ -2303,7 +2425,7 @@ show_startup_msgs (whois_session_params * wsess)
                 else fprintf(stderr,"Using Prefix WhoIs for bulk file resolution.\n");
             }
         }
-        if (show_routes_byprefix == 1 && show_routes_byasn == 1) {
+        if (show_routes_byprefix == 1 && (show_routes_byasn == 1 || show_routes_bytransitasn == 1)) {
             fprintf(stderr,"You may only perform routeviews one at a time.  Using by-ASN.\n");
             show_routes_byprefix = 0;
         } 
@@ -2324,28 +2446,38 @@ int main(int ac, char **av)
     
     wsess = w_init();
     setbuf(stdout, NULL);
-    
-    while ((ch = getopt (ac, av, "aCcfgNnOopPRrstuVvh:w:")) != EOF)
+        
+    while ((ch = getopt (ac, av, "AaCcfgNnOopPRrstuVvh:w:")) != EOF)
         switch (ch) {
             case 'a':
                 use_gigo = 0;
                 show_routes_byasn = 1;
+                go_interactive = 1;
+                break;
+            case 'A':
+                use_gigo = 0;
+                show_routes_bytransitasn = 1;
+                go_interactive = 1;
                 break;
             case 'N':
                 use_gigo = 0;
                 show_networks_byasn = 1;
+                go_interactive = 1;
                 break;
             case 'O':
                 use_gigo = 0;
                 show_contacts_byasn = 1;
+                go_interactive = 1;
                 break;
             case 's':
                 use_gigo = 0;
                 show_server_status = 1;
+                go_interactive = 1;
                 break;
             case 'P':
                 use_gigo = 0;
                 show_routes_byprefix = 1;
+                go_interactive = 1;
                 break;
             case 'v':
                 usage(av[0]);
@@ -2409,6 +2541,7 @@ int main(int ac, char **av)
                 break;
             case 'n':
                 display_netname = 1;
+                go_interactive = 1;
                 break;
             case 'r':
                 use_riswhois = 1;
@@ -2416,12 +2549,15 @@ int main(int ac, char **av)
                 break;
             case 'R':
                 display_radb_as = 1;
+                go_interactive = 1;
                 break;
             case 'o':
                 display_orgname = 0;
+                go_interactive = 1;
                 break;
             case 'p':
                 display_aspath = 1;
+                go_interactive = 1;
                 break;
             case 'f':
                 use_gigo = 0;
@@ -2432,23 +2568,62 @@ int main(int ac, char **av)
                 break;
             case 'g':
                 use_gigo = 0;
+                go_interactive = 1;
                 break;
             default:
                 usage (av[0]);
         }
+
+    /* Catch hostname input without any arguments */
+    if ((ac - optind) > 0) 
+        go_interactive = 1;
+                            
+    if (go_interactive > 0) {
+    /*  Quickly check stdin for input to avoid David having to type   */
+    /*  '-f -' to pipe into to whob, even if it has args :-)          */
+        fd_set rfds;
+        struct timeval timev;
+        int selretval;
+     
+        /* Watch stdin (fd 0) briefly for input. */
+        FD_ZERO(&rfds);
+        FD_SET(0, &rfds);
+        
+        timev.tv_sec = 0;
+        timev.tv_usec = 100;
+     
+        selretval = select(1, &rfds, NULL, NULL, &timev);
+     
+        if (selretval) {
+            /* There's input in them there hills, treat it as the input file */
+            use_gigo = 0; 
+            read_from_file = 1;
+            use_stdin = 1;
+        } 
+        /*
+          else
+            if (wsess->w_noisy) fprintf(stderr,"No input found on stdin.\n");
+        */
+    } else if (read_from_file < 1) {
+        /* No indication we should be interactive based on arguments, so wait for STDIN */
+        use_gigo = 0;
+        read_from_file = 1;
+        use_stdin = 1;
+    }
             
-            if (((ac - optind) < 1) && (show_server_status != 1))
-                usage (av[0]);
-    
+    if (((ac - optind) < 1) && (show_server_status != 1) && (go_interactive > 0))
+        usage (av[0]);
+        
     /* Show the verbose startup information if verbosity is enabled/requested */
     show_startup_msgs(wsess);
-    
+                    
     if (show_server_status == 1) {
         w_display_pwhois_version(wsess);
         exit(EXIT_FAILURE);
-    } else { 
-        if (strlen(av[optind]) > max_hostname_input) {
-            fprintf(stderr,"Sorry, the host name you supplied was unreasonably long.\n");
+    } else if ((read_from_file > 0 && use_stdin < 1) || go_interactive > 0) { 
+    
+        if (strlen(av[optind]) > max_hostname_input) {        
+            fprintf(stderr,"Sorry, the subject name you supplied was unreasonably long.\n");
             exit(EXIT_FAILURE);
         } else {
             strncpy(hostname,av[optind],strlen(av[optind]));
@@ -2467,16 +2642,27 @@ int main(int ac, char **av)
     }
     
     if (use_gigo > 0) {
-        w_display_pwhois_gigo(wsess, hostname);
-        exit(EXIT_SUCCESS);
+
+        if (strncmp(hostname,"me",2) == 0 || strncmp(hostname,"whoami",6) == 0) {
+            w_display_myaddress(wsess);
+            exit(EXIT_SUCCESS);
+        } else {
+           w_display_pwhois_gigo(wsess, hostname);
+           exit(EXIT_SUCCESS);
+	}
     }
     
-    if ((show_routes_byasn || show_contacts_byasn || show_networks_byasn) && (strlen(hostname) <= 10) && atoi(hostname)) {
+    if ((show_routes_byasn || show_routes_bytransitasn || show_contacts_byasn || show_networks_byasn) && (strlen(hostname) <= 10) && atoi(hostname) && stricontains(hostname,".") < 1) {
         user_asn = atoi(hostname);
-        if (wsess->w_noisy > 1) printf("Using user-supplied ASN %d for lookup.\n",user_asn);
+        if (wsess->w_noisy > 1) fprintf(stderr,"Using user-supplied ASN %d for lookup.\n",user_asn);
     } else if (show_routes_byprefix == 1) {
         printf("Displaying all routes for prefix %s.\n",hostname);
     } else {
+
+	if (strncmp(hostname,"me",2) == 0 || strncmp(hostname,"whoami",6) == 0) {
+	    w_display_myaddress(wsess);
+	    exit(EXIT_SUCCESS);
+	}
         
         if (inet_aton(hostname, &in)) {
             addr = hostname;
@@ -2504,6 +2690,20 @@ int main(int ac, char **av)
                 printf("Sorry, unable to resolve the ASN for %s (%s) at this time.\n",hostname,addr);
             }
         }
+    } else if (show_routes_bytransitasn) {
+        if (user_asn > 0) {
+            printf("Displaying all routes whose Transit-AS is %d.\n", user_asn);
+            snprintf(user_asn_buf,9,"%d",user_asn);
+            w_display_rvbytransitasn_pwhois(wsess, user_asn_buf);
+        } else {
+            w_lookup_all_pwhois(wsess, addr);
+            if (atoi(wsess->consolidated_asn)) {
+                printf("Displaying all routes whose Transit-AS is %s.\n", wsess->consolidated_asn);
+                w_display_rvbytransitasn_pwhois(wsess, wsess->consolidated_asn);
+            } else {
+                printf("Sorry, unable to resolve the ASN for %s (%s) at this time.\n",hostname,addr);
+            }
+        }
     } else if (show_networks_byasn) {
         if (user_asn > 0) {
             printf("Displaying all networks registered to the Origin-AS %d.\n", user_asn);
